{$A-,B-,D-,E-,F+,G+,I-,L-,N-,O-,R-,S-,V-}


UNIT extragraph;

(**********************************************************************)
(**********************************************************************)
(***                                                                ***)
(*** Diese Unit stellt Funktionen zur Verfgung, die die von Graph  ***)
(*** ergnzt bzw. deren Handhabung vereinfacht. Als Datenstruktu-   ***)
(*** ren werden Area und Shape angeboten. Area dient zum Speichern  ***)
(*** von rechteckigen Bildschirmausschnitten, die bei Bedarf wieder ***)
(*** restauriert werden knnen. Damit ist es mglich, einen Teil    ***)
(*** des Bildschirms zu retten, ihn zu bearbeiten, und nachher wie- ***)
(*** der in die ursprngliche Form zu bringen. Fenster und Menue-   ***)
(*** techniken werden darauf aufgebaut. Shape dient zum Verwalten   ***)
(*** von Formen, die Bitmig erstellt ber den Grafikschirm wan-   ***)
(*** dern knnen. Hiermit knnen Spiele und sonstige bewegte Gra-   ***)
(*** fiken programmiert werden.                                     ***)
(***                                                                ***)
(**********************************************************************)
(**********************************************************************)

INTERFACE


USES Graph,Crt,DOS;


CONST maxGraphSave=10;           (*** grte Tiefe des Parame-  ***)
                                 (*** terstapels                ***)

      noVGA=0;
      VGA16=1;
      VGA256=2;


TYPE Area=RECORD                 (*** dient zum Speichern eines ***)
              x,y,sx,sy:WORD;    (*** Bitmaps mit SaveArea und  ***)
              ysplit:ARRAY[0..16] OF WORD; (*** RestoreArea     ***)
              p:ARRAY[0..15] OF POINTER;
              size:ARRAY[0..15] OF LONGINT
          END;

     Shape=RECORD                (*** dient zum Speichern eines ***)
               sx,sy:WORD;       (*** Bitmaps das bequem ber   ***)
               restore:BOOLEAN;  (*** Bildschirm bewegt werden  ***)
               a:Area;           (*** kann                      ***)
               p:POINTER;
               x,y:WORD
           END;

     ScreenArray=ARRAY[0..32767] OF BYTE;

     RGBColor=RECORD
        r,g,b:BYTE    (* Alle von 0 bis 63 *)
     END;

     ColorPalette=ARRAY[0..63,0..2] OF BYTE;     (*** Farbpalette mit 64 Farben  ***)

     ColorPalette256=ARRAY[0..255,0..2] OF BYTE; (*** Farbpalette mit 256 Farben ***)


VAR IORG:WORD;                     (*** Speichert Wert von IOResult       ***)


    HerculesScreen:ScreenArray ABSOLUTE $B000:$0000;
    Vga1Screen:ScreenArray ABSOLUTE $A000:$0000;
    Vga2Screen:ScreenArray ABSOLUTE $A800:$0000;

    GraphSaveList:ARRAY[1..maxGraphSave] OF RECORD  (*** Graphikeinstellungen ***)
                                      TextSave:TextSettingsType;
                                      FillSave:FillSettingsType;
                                      LineSave:LineSettingsType;
                                      FillPatternSave:FillPatternType;
                                      BkColor:WORD;
                                      Color:WORD
                                  END;
    GraphSaveTop:BYTE;


FUNCTION ExtraGraphResult:BYTE;
(*** Gibt Fehler zurck und setzt internen Fehler-  ***)
(*** status auf 0. Mgliche Fehler:                 ***)
(***    0:alles OK                                  ***)
(***    1:nicht genug Speicher vorhanden            ***)


PROCEDURE SetGraphDriver(Driver,Mode:INTEGER);
(*** Spezifiziert Graphiktreiber und Modus. Wird    ***)
(*** Diese Prozedur nicht aufgerufen, so wird der   ***)
(*** Treiber ausgewhlt.                            ***)


PROCEDURE SetUserGraphDriver(DriverName:STRING;Mode:INTEGER);
(*** Spezifiziert Usergraphiktreiber und Modus      ***)


PROCEDURE InitGraphScreen;
(*** Ldt und initialisiert Graphiktreiber          ***)


PROCEDURE InitGraphScreenPath(path:STRING);
(*** Ldt und initialisiert Graphiktreiber, der     ***)
(*** an der angegebenen Stelle im System liegt      ***)


PROCEDURE GraphScreen;
(*** Schaltet auf Grafikbildschirm um und sorgt da- ***)
(*** fr, da auch der Maustreiber davon erfhrt    ***)
(*** wenn einer geladen ist                         ***)


FUNCTION VgaScreen:BOOLEAN;
(*** Ist wahr, wenn EGA- oder VGA-Modus eingeschal- ***)
(*** tet ist.                                       ***)


PROCEDURE TextScreen;
(*** Schaltet auf Textschirm um, Maustreiber wird   ***)
(*** wieder auf alten Zustand gebracht              ***)


PROCEDURE ClearGraphScreen;
(*** Lscht den eingeschalteten Grafikschirm        ***)


PROCEDURE SaveGraphScreen(name:STRING;vga:BOOLEAN);
(*** Speichert Hercules oder VGA Bildschirm auf     ***)
(*** externem Datentrger unter angegebenem Namen.  ***)
(*** IORG wird dabei gesetzt                        ***)


PROCEDURE LoadGraphScreen(name:STRING;vga:BOOLEAN);
(*** Zeigt extern gespeicherten Bildschirm wieder  ***)
(*** an. IORG wird gesetzt                         ***)


FUNCTION OrgX(x:INTEGER):INTEGER;
FUNCTION OrgY(y:INTEGER):INTEGER;
(*** wandelt Hercules Koordinaten auf die aktuell   ***)
(*** eingestellte Grafikkarte um                    ***)


FUNCTION BitsPerPixel:BYTE;
(*** Gibt an, wieviele Bits zur Speicherung eines   ***)
(*** Pixels im  eingestellten Graphikmodus ntig    ***)
(*** sind.                                          ***)


PROCEDURE SavePlane;
(*** Rettet die VGA-Planeregister.                  ***)


PROCEDURE RestorePlane;
(*** Holt die VGA-Planeregister zurck.             ***)


PROCEDURE SetPlane(level:BYTE);
(*** Setzt Plane bei EGA- oder VGA-Karte. level ist ***)
(*** dabei im Bereich 0..3.                         ***)


PROCEDURE GetVGAPalette(color:BYTE;VAR red,green,blue:BYTE);
(*** Gibt Palette fr angegebene Farbe zurck.      ***)


PROCEDURE SetVGAPalette(color,red,green,blue:BYTE);
(*** Setzt Palette fr angegebene Farbe.            ***)


PROCEDURE GetAllVGAPalette(VAR pal:ColorPalette);
(*** Gibt alle 64 Farbeintrge auf einmal zurck.   ***)


PROCEDURE SetAllVGAPalette(VAR pal:ColorPalette);
(*** Setzt alle 64 Farbeintrge auf einmal.         ***)


PROCEDURE GetAllVGAPalette256(VAR pal:ColorPalette256);
(*** Gibt alle 256 Farbeintrge auf einmal zurck.  ***)


PROCEDURE SetAllVGAPalette256(VAR pal:ColorPalette256);
(*** Setzt alle 256 Farbeintrge auf einmal.        ***)


PROCEDURE SetOverScan(color:BYTE);
(*** Setzt das Overscanregister auf die gewnschte  ***)
(*** Farbe.                                         ***)


PROCEDURE ClearArea(x,y,sx,sy:WORD);
(*** Lscht den angegebenen Bereich von (x,y) bis   ***)
(*** (x+sx,y+sy)                                    ***)


PROCEDURE SaveArea(VAR a:Area;x,y,sx,sy:WORD);
(*** Speichert angegebenen Bereich in der Variablen ***)
(*** vom Typ Area                                   ***)


PROCEDURE RestoreArea(VAR a:Area);
(*** Holt den zuvor mit SaveArea gespeicherten      ***)
(*** Bildbereich wieder auf den Grafikschirm        ***)


PROCEDURE FreeArea(VAR a:Area);
(*** Gibt den Speicher des Bildes wieder frei, ohne ***)
(*** es anzuzeigen.                                 ***)


PROCEDURE MoveArea(VAR a:Area;x,y:WORD);
(*** Holt den Bildbereich an eine andere Stelle     ***)


PROCEDURE CopyArea(VAR a:Area;x,y:WORD);
(*** Wie MoveArea, lscht jedoch nicht Orginal.     ***)


PROCEDURE SaveAreaOnDisk(VAR a:Area;name:STRING);
(*** Speichert ein Area auf Disk                    ***)


PROCEDURE LoadAreaFromDisk(VAR a:Area;name:STRING);
(*** Holt Area von Diskette                         ***)


PROCEDURE Rect(x,y,sx,sy:WORD);
(*** Zeichnet ein Rechteck an x,y als linke obere   ***)
(*** Ecke. sx,sy geben die Gre an                 ***)


PROCEDURE RoundRect(x,y,sx,sy,rad:WORD);
(*** Zeichnet ein Rechteck mit runden Ecken an x,y  ***)
(*** als linke obere Ecke. sx,sy geben die Gre    ***)
(*** an. rad gibt den Radius der Eckenkreise an     ***)


PROCEDURE SetCol(col:BYTE);
(*** Setzt ein Pattern mit einem Grauwert von 0     ***)
(*** bis 7 (analog zu SetFillPattern)               ***)


PROCEDURE SetColCol(col1,col2:BYTE);
(*** Setzt ein Pattern mit Grauwert col2 (0..7) und ***)
(*** Farbe col1.                                    ***)

FUNCTION Color(col:BYTE):BYTE;
(*** Gibt eine 1 zurck, wenn keine Farbgrafik ein- ***)
(*** geschaltet ist, sonst wird der Wert unvern-   ***)
(*** dert bernommen.                               ***)


PROCEDURE PutPixelCol(x,y:INTEGER;col:BYTE);
(*** Simuliert Punktsetzen mit Farben auf Monochrom-***)
(*** Adaptern indem PutPixel nur mit der dem Grau-  ***)
(*** Wert entsprechenden Hufigkeit ausgefhrt wird.***)
(*** Der Grauwert entspricht dem von SetCol.        ***)


PROCEDURE MakeShape(VAR sh:Shape;
                    sx,sy:WORD;
                    vga:BYTE;
                    restore:BOOLEAN);
(*** Legt den Speicherplatz fr ein Shape an und    ***)
(*** setzt Grenparameter. restore gibt an, ob der ***)
(*** Untergrund gespeichert werden soll und bei ei- ***)
(*** ner Lschung oder Bewegung wieder restauriert  ***)
(*** werden soll.                                   ***)
(*** vga hat den Wert noVGA,VGA16 oder VGA256       ***)


PROCEDURE MakeShapeWithCol(VAR sh:Shape;
                    sx,sy:WORD;
                    vga:BYTE;
                    restore:BOOLEAN;
                    c1,c2:BYTE);
(*** Wie oben, nur da man 2 Farben angeben kann.   ***)


PROCEDURE DefineShape(VAR sh:Shape;s:STRING);
(*** Legt das Punktmuster eines Shapes fest. Im     ***)
(*** String steht fr jeden nichtgesetzten Punkt    ***)
(*** ein Blank sonst ein Nicht-Blank. Vor dem Auf-  ***)
(*** ruf von DefineShape mu MakeShape erfolgt      ***)
(*** sein.                                          ***)


FUNCTION ShapeSize(sx,sy:WORD;vga:BYTE):WORD;
(*** Entspricht ImageSize, jedoch nur fr Hercules  ***)
(*** und VGA.                                       ***)
(*** vga hat den Wert noVGA,VGA16 oder VGA256       ***)


PROCEDURE ShowShape(VAR sh:Shape;x,y:INTEGER;Mode:WORD);
(*** Zeigt definiertes Shape an spezifizierter Ko-  ***)
(*** ordinate. Mode gibt die Verknpfung mit dem    ***)
(*** Untergrund an:                                 ***)
(***                0:berschreiben                 ***)
(***                1:Xor mit Screen                ***)
(***                2:Or mit Screen                 ***)
(***                3:And mit Screen                ***)
(***                4:invertiert berschreiben      ***)


PROCEDURE HideShape(VAR sh:Shape);
(*** Lscht Shape und restauriert Untergrund (nur   ***)
(*** mglich wenn restore=TRUE bei MakeShape)       ***)


PROCEDURE MoveShape(VAR sh:Shape;x,y:INTEGER;Mode:WORD);
(*** Wie ShowShape, nur da das Shape auf der al-   ***)
(*** ten Position gelscht wird (nur mglich, wenn  ***)
(*** restore=TRUE bei MakeShape).                   ***)


FUNCTION StyleCorrect(style:WORD):REAL;
(*** Gibt einen Korrekturfaktor an, der die ver-    ***)
(*** schiedenen Schriftarten auf eine Gre konver- ***)
(*** tiert.                                         ***)


PROCEDURE SaveGraphParams;
(*** Speichert alle relevanten Graphikeinstellungen.***)
(*** Diese Routine kann mit RestoreGraphParams ge-  ***)
(*** schachtelt aufgerufen werden.                  ***)


PROCEDURE RestoreGraphParams;
(*** Holt Graphikeinstellungen zurck.              ***)


PROCEDURE MakeBlackWhitePalette;
(*** Macht eine Palette mit 256 Grauwerten.         ***)


PROCEDURE MakeDithPalette;
(*** Stellt die Palette zur Verfgung, die man beim ***)
(*** Dithering braucht.                             ***)

PROCEDURE MakeRedGreenPalette;
(*** Stellt die Palette zur Verfgung, die man fr  ***)
(*** Rot-Grn-Holo braucht.                         ***)


PROCEDURE DithPutPixel(x,y:INTEGER;Red,Green,Blue:INTEGER);
(*** Setzt einen Punkt mit den Farbwerten Red,Blue  ***)
(*** und Green (aus 0..63). Da nur 256 Farben       ***)
(*** gleichzeitig angezeigt werden knnen, wird die ***)
(*** Farbe u.u. mit Dithering angenhert. Daher mu ***)
(*** die Palette mit MakePalette initialisiert wer- ***)
(*** den.                                           ***)


PROCEDURE RGPutPixel(x,y:INTEGER;Red,Green:INTEGER);
(*** Setzt Punkte fr Rot-Grn-Holo.                ***)


PROCEDURE SetDithColor(r,g,b:INTEGER);
(*** Setzt die Farbe fr z.B. DithLine.             ***)


PROCEDURE DithLine(x0,y0,x1,y1:INTEGER);
(*** Zieht eine Linie mit Ditherfarben.             ***)


PROCEDURE OutTextXYNumb(x,y:INTEGER;s:String;nr:INTEGER);
(*** Wie OutTextXY nur mit String und INTEGER;      ***)


IMPLEMENTATION

CONST colors:ARRAY[0..7] OF FillPatternType=
                 (($00,$00,$00,$00,$00,$00,$00,$00),
                  ($00,$22,$00,$00,$00,$22,$00,$00),
                  ($00,$44,$00,$11,$00,$44,$00,$11),
                  ($00,$55,$00,$55,$00,$55,$00,$55),
                  ($AA,$55,$AA,$55,$AA,$55,$AA,$55),
                  ($FF,$AA,$FF,$AA,$FF,$AA,$FF,$AA),
                  ($FF,$CC,$FF,$EE,$FF,$CC,$FF,$EE),
                  ($FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF));


VAR graphmode:BYTE ABSOLUTE$0040:$0049;
    oldmode:BYTE;
    shapecount:WORD;
    vgaflag:BYTE;
    GrafDriver,GrafMode:INTEGER;
    GraphR:WORD;
    Port3c4,Port3c5,Port3ce,Port3cf:BYTE;
    DithColor:RGBColor;
    col1,col2:BYTE;


FUNCTION ExtraGraphResult:BYTE;
BEGIN
    ExtraGraphResult:=GraphR;
    GraphR:=0
END;


PROCEDURE GraphError(nr:BYTE);
BEGIN
    TextScreen;
    WriteLn('Extragraph error:');
    CASE nr OF
        1:WriteLn('too many "SaveGraphParams"');
        2:WriteLn('"RestoreGraphParams" without "SaveGraphParams"');
     ELSE WriteLn('unknown Error')
    END;
    WriteLn;
    Halt(1)
END;


PROCEDURE SetGraphDriver(Driver,Mode:INTEGER);
BEGIN
    GrafDriver:=Driver;
    GrafMode:=Mode
END;


PROCEDURE SetUserGraphDriver(DriverName:STRING;Mode:INTEGER);
BEGIN
    GrafDriver:=InstallUserDriver(DriverName,NIL)+5;
    GrafMode:=Mode
END;


PROCEDURE InitGraphScreen;
BEGIN
    InitGraph(GrafDriver,GrafMode,'')
END;


PROCEDURE InitGraphScreenPath(path:STRING);
BEGIN
    InitGraph(GrafDriver,GrafMode,path)
END;


PROCEDURE GraphScreen;
BEGIN
    SetGraphMode(Grafmode);
    oldmode:=graphmode;
    IF GrafDriver=HercMono THEN
        graphmode:=6
    ELSE IF GetMaxX>639 THEN
             graphmode:=18
END;


FUNCTION VgaScreen:BOOLEAN;
BEGIN
    VgaScreen:=(SUCC(GetMaxX) AND 31=0)
END;


PROCEDURE TextScreen;
BEGIN
    graphmode:=oldmode;
    RestoreCrtMode
END;


PROCEDURE ClearGraphScreen;
BEGIN
    ClearArea(0,0,GetMaxX,GetMaxY)
END;


PROCEDURE SaveGraphScreen(name:STRING;vga:BOOLEAN);
VAR pict:FILE OF ScreenArray;
    i:BYTE;
BEGIN
    Assign(pict,name);
    (*$I-*)
    ReWrite(pict);
    (*$I+*)
    IORG:=IOResult;
    IF IORG=0 THEN
        IF vga THEN
        BEGIN
            SavePlane;
            FOR i:=0 TO 3 DO
            BEGIN
                SetPlane(i);
                Write(pict,Vga1Screen);
                Write(pict,Vga2Screen)
            END;
            RestorePlane;
            Close(pict)
        END
        ELSE BEGIN
            Write(pict,HerculesScreen);
            Close(pict)
        END
END;


PROCEDURE LoadGraphScreen(name:STRING;vga:BOOLEAN);
VAR pict:FILE OF ScreenArray;
    i:BYTE;
BEGIN
    Assign(pict,name);
    (*$I-*)
    Reset(pict);
    (*$I+*)
    IORG:=IOResult;
    IF IORG=0 THEN
        IF vga THEN
        BEGIN
            SavePlane;
            FOR i:=0 TO 3 DO
            BEGIN
                SetPlane(i);
                Read(pict,Vga1Screen);
                Read(pict,Vga2Screen)
            END;
            RestorePlane;
            Close(pict)
        END
        ELSE BEGIN
            Read(pict,HerculesScreen);
            Close(pict)
        END
END;


FUNCTION OrgX(x:INTEGER):INTEGER;
VAR l:LONGINT;
BEGIN
    l:=x*GetMaxX;
    OrgX:=l DIV 720
END;


FUNCTION Orgy(y:INTEGER):INTEGER;
VAR l:LONGINT;
BEGIN
    l:=y*GetMaxY;
    OrgY:=l DIV 720
END;


FUNCTION BitsPerPixel:BYTE;
BEGIN
    BitsPerPixel:=(ImageSize(0,0,31,31)-4)*8 DIV 1024
END;


PROCEDURE SavePlane;
BEGIN
    Port3c4:=Port[$3C4];
    Port[$3c4]:=2;
    Port3c5:=Port[$3C5];
    Port3ce:=Port[$3CE];
    Port[$3CE]:=4;
    Port3cf:=Port[$3CF]
END;


PROCEDURE RestorePlane;
BEGIN
    Port[$3C4]:=2;
    Port[$3C5]:=Port3c5;
    Port[$3C4]:=Port3c4;
    Port[$3CE]:=4;
    Port[$3CF]:=Port3cf;
    Port[$3CE]:=Port3ce
END;


PROCEDURE SetPlane(level:BYTE);
BEGIN
    Port[$3C4]:=2;
    Port[$3C5]:=1 SHL level OR (Port[$3C5] AND 240);
    Port[$3CE]:=4;
    Port[$3CF]:=Port[$3CF] AND 252 OR level
END;


PROCEDURE GetVGAPalette(color:BYTE;VAR red,green,blue:BYTE);
VAR r:REGISTERS;
BEGIN
    r.AX:=$1015;
    r.BX:=color;
    Intr($10,r);
    red:=r.DH;
    green:=r.CH;
    blue:=r.CL
END;


PROCEDURE SetVGAPalette(color,red,green,blue:BYTE);
VAR r:REGISTERS;
BEGIN
    r.AX:=$1010;
    r.BX:=color;
    r.DH:=red;
    r.CH:=green;
    r.CL:=blue;
    Intr($10,r)
END;


PROCEDURE GetAllVGAPalette(VAR pal:ColorPalette);
VAR r:REGISTERS;
BEGIN
    r.AX:=$1017;
    r.ES:=Seg(pal);
    r.DX:=Ofs(pal);
    r.BX:=0;
    r.CX:=64;
    Intr($10,r)
END;


PROCEDURE SetAllVGAPalette(VAR pal:ColorPalette);
VAR r:REGISTERS;
BEGIN
    r.AX:=$1012;
    r.ES:=Seg(pal);
    r.DX:=Ofs(pal);
    r.BX:=0;
    r.CX:=64;
    Intr($10,r)
END;


PROCEDURE GetAllVGAPalette256(VAR pal:ColorPalette256);
VAR r:REGISTERS;
BEGIN
    r.AX:=$1017;
    r.ES:=Seg(pal);
    r.DX:=Ofs(pal);
    r.BX:=0;
    r.CX:=256;
    Intr($10,r)
END;


PROCEDURE SetAllVGAPalette256(VAR pal:ColorPalette256);
VAR r:REGISTERS;
BEGIN

    r.AX:=$1012;
    r.ES:=Seg(pal);
    r.DX:=Ofs(pal);
    r.BX:=0;
    r.CX:=256;
    Intr($10,r)
END;


PROCEDURE SetOverScan(color:BYTE);
VAR r:REGISTERS;
BEGIN
    r.AX:=$1001;
    r.BH:=color;
    Intr($10,r)
END;


PROCEDURE ClearArea(x,y,sx,sy:WORD);
VAR OldColor:WORD;
BEGIN
   IF x>32767 THEN
      x:=0;
   IF y>32767 THEN
      y:=0;
   SetFillStyle(0,0);
   OldColor:=GetColor;
   SetColor(GetBkColor);
   Bar(x,y,x+sx,y+sy);
   SetColor(OldColor)
END;


CONST EndArea=65535;


PROCEDURE SaveArea(VAR a:Area;x,y,sx,sy:WORD);
VAR aktsy:WORD;
    i:BYTE;
BEGIN
    a.x:=x;
    a.y:=y;
    a.sx:=sx;
    a.sy:=sy;
    i:=0;
    WHILE sy>0 DO
    BEGIN
        a.ysplit[i]:=y;
        aktsy:=sy;
        IF aktsy>150 THEN
            aktsy:=150;
        a.size[i]:=ImageSize(x,y,x+sx,y+aktsy);
        IF MaxAvail<a.size[i] THEN
        BEGIN
            WHILE i>0 DO
            BEGIN
                DEC(i);
                FreeMem(a.p[i],a.size[i])
            END;
            GraphR:=1;
            Exit
        END;
        GetMem(a.p[i],a.size[i]);
        GetImage(x,y,x+sx,y+aktsy,a.p[i]^);
        y:=y+aktsy;
        sy:=sy-aktsy;
        INC(i);
    END;
    a.ysplit[i]:=EndArea;
END;


PROCEDURE RestoreArea(VAR a:Area);
VAR i:BYTE;
BEGIN
    i:=0;
    WHILE a.ysplit[i]<>EndArea DO
    BEGIN
        PutImage(a.x,a.ysplit[i],a.p[i]^,NormalPut);
        FreeMem(a.p[i],a.size[i]);
        INC(i)
    END
END;


PROCEDURE FreeArea(VAR a:Area);
VAR i:BYTE;
BEGIN
    i:=0;
    WHILE a.ysplit[i]<>EndArea DO
    BEGIN
        FreeMem(a.p[i],a.size[i]);
        INC(i)
    END
END;


PROCEDURE MoveArea(VAR a:Area;x,y:WORD);
VAR i:BYTE;
BEGIN
    i:=0;
    WHILE a.ysplit[i]<>EndArea DO
    BEGIN
        PutImage(x,a.ysplit[i]+y-a.y,a.p[i]^,NormalPut);
        FreeMem(a.p[i],a.size[i]);
        INC(i)
    END
END;


PROCEDURE CopyArea(VAR a:Area;x,y:WORD);
VAR i:BYTE;
BEGIN
    i:=0;
    WHILE a.ysplit[i]<>EndArea DO
    BEGIN
        PutImage(x,a.ysplit[i]+y-a.y,a.p[i]^,NormalPut);
        INC(i)
    END
END;


PROCEDURE SaveAreaOnDisk(VAR a:Area;name:STRING);
VAR f:FILE;
    i:BYTE;
BEGIN
    Assign(f,name);
    (*$I-*)
    ReWrite(f,1);
    (*$I+*)
    IORG:=IOResult;
    IF IORG=0 THEN
    BEGIN
        BlockWrite(f,a,SizeOf(a));
        i:=0;
        WHILE a.ysplit[i]<>EndArea DO
        BEGIN
            BlockWrite(f,a.p[i]^,a.size[i]);
            INC(i)
        END;
        Close(f)
    END
END;


PROCEDURE LoadAreaFromDisk(VAR a:Area;name:STRING);
VAR f:FILE;
    i:BYTE;
BEGIN
    Assign(f,name);
    (*$I-*)
    Reset(f,1);
    (*$I+*)
    IORG:=IOResult;
    IF IORG=0 THEN
    BEGIN
        BlockRead(f,a,SizeOf(a));
        i:=0;
        WHILE a.ysplit[i]<>EndArea DO
        BEGIN
            IF MaxAvail<a.size[i] THEN
            BEGIN
                WHILE i>0 DO
                BEGIN
                    DEC(i);
                    FreeMem(a.p[i],a.size[i])
                END;
                GraphR:=1;
                Close(f);
                Exit
            END;
            GetMem(a.p[i],a.size[i]);
            BlockRead(f,a.p[i]^,a.size[i]);
            INC(i)
        END;
        Close(f)
    END
END;


PROCEDURE Rect(x,y,sx,sy:WORD);
BEGIN
    Rectangle(x,y,x+sx,y+sy)
END;


PROCEDURE RoundRect(x,y,sx,sy,rad:WORD);
VAR aspx,aspy,rady:WORD;
BEGIN
    GetAspectRatio(aspx,aspy);
    rady:=ROUND(aspx/aspy*rad);
    MoveTo(x,y+rady);
    LineTo(x,y+sy-rady);
    MoveTo(x+sx,y+rady);
    LineTo(x+sx,y+sy-rady);
    MoveTo(x+rad,y);
    Lineto(x+sx-rad,y);
    MoveTo(x+rad,y+sy);
    Lineto(x+sx-rad,y+sy);
    Arc(x+rad,y+rady,90,180,rad);
    Arc(x+rad,y+sy-rady,180,270,rad);
    Arc(x+sx-rad,y+rady,0,90,rad);
    Arc(x+sx-rad,y+sy-rady,270,360,rad)
END;


PROCEDURE SetCol(col:BYTE);
BEGIN
    SetFillPattern(colors[col],1)
END;


PROCEDURE SetColCol(col1,col2:BYTE);
BEGIN
    SetFillPattern(colors[col2],col1)
END;


FUNCTION Color(col:BYTE):BYTE;
BEGIN
    IF (col<>0) AND (GetMaxColor=1) THEN
        Color:=1
    ELSE
        Color:=col
END;


PROCEDURE PutPixelCol(x,y:INTEGER;col:BYTE);
VAR xm,ym:BYTE;
BEGIN
    xm:=x AND 3;
    ym:=y AND 3;
    CASE col OF
        0:PutPixel(x,y,0);
        1:IF xm*ym=1 THEN
              PutPixel(x,y,1)
          ELSE
              PutPixel(x,y,0);
        2:IF xm*ym AND 7=1 THEN
              PutPixel(x,y,1)
          ELSE
              PutPixel(x,y,0);
        3:IF xm*ym AND 1=1 THEN
              PutPixel(x,y,1)
          ELSE
              PutPixel(x,y,0);
        4:IF (xm+ym) AND 1=0 THEN
              PutPixel(x,y,1)
          ELSE
              PutPixel(x,y,0);
        5:IF xm*ym AND 1=0 THEN
              PutPixel(x,y,1)
          ELSE
              PutPixel(x,y,0);
        6:IF xm*ym<>1 THEN
              PutPixel(x,y,1)
          ELSE
              PutPixel(x,y,0);
        7:PutPixel(x,y,1)
    END
END;


PROCEDURE MakeShape(VAR sh:Shape;
                    sx,sy:WORD;
                    vga:BYTE;
                    restore:BOOLEAN);

TYPE atype=ARRAY[0..1] OF WORD;

VAR p2:^atype;
    size:WORD;

BEGIN
    sh.sx:=sx;
    sh.sy:=sy;
    sh.restore:=restore;
    size:=ShapeSize(sx,sy,vga);
    IF MaxAvail<size THEN
        GraphR:=1
    ELSE BEGIN
        GetMem(sh.p,size);
        p2:=sh.p;
        p2^[0]:=sx-1;
        p2^[1]:=sy-1;
        shapecount:=3;
        vgaflag:=vga
    END;
    col1:=16;
    col2:=0
END;


PROCEDURE MakeShapeWithCol(VAR sh:Shape;
                    sx,sy:WORD;
                    vga:BYTE;
                    restore:BOOLEAN;
                    c1,c2:BYTE);

TYPE atype=ARRAY[0..1] OF WORD;

VAR p2:^atype;
    size:WORD;

BEGIN
    sh.sx:=sx;
    sh.sy:=sy;
    sh.restore:=restore;
    size:=ShapeSize(sx,sy,vga);
    IF MaxAvail<size THEN
        GraphR:=1
    ELSE BEGIN
        GetMem(sh.p,size);
        p2:=sh.p;
        p2^[0]:=sx-1;
        p2^[1]:=sy-1;
        shapecount:=3;
        vgaflag:=vga
    END;
    col1:=c1;
    col2:=c2
END;



PROCEDURE DefineShape(VAR sh:Shape;s:STRING);

TYPE atype=ARRAY[0..2047] OF BYTE;

VAR p2:^atype;
    i,j:BYTE;
    anz:BYTE;

BEGIN
    p2:=sh.p;
    IF vgaflag=VGA256 THEN
    BEGIN
        FOR i:=1 TO Length(s) DO
        BEGIN
            INC(shapecount);
            IF s[i]<>' ' THEN
                p2^[shapecount]:=col1
            ELSE
                p2^[shapecount]:=col2;
        END
    END
    ELSE BEGIN
        anz:=1;
        IF vgaflag=VGA16 THEN
            anz:=4;
        FOR j:=1 TO anz DO
        BEGIN
            FOR i:=1 TO Length(s) DO
            BEGIN
                IF PRED(i) MOD 8=0 THEN
                BEGIN
                    INC(shapecount);
                    p2^[shapecount]:=0
                END;
                p2^[shapecount]:=p2^[shapecount]*2+ORD(s[i]<>' ')
            END;
            FOR i:=(Length(s)+7) MOD 8 TO 6 DO
                p2^[shapecount]:=p2^[shapecount]*2;
            IF p2^[0] MOD 8=1 THEN
                INC(shapecount)
        END
    END
END;


FUNCTION ShapeSize(sx,sy:WORD;vga:BYTE):WORD;
BEGIN
    CASE vga OF
        noVGA:ShapeSize:=(sx DIV 8)*sy+6;
        VGA16:ShapeSize:=(sx DIV 8)*sy*4+6;
        VGA256:ShapeSize:=(sx DIV 8)*sy*8+6;
    END
END;


PROCEDURE ShowShape(VAR sh:Shape;x,y:INTEGER;Mode:WORD);
BEGIN
    IF sh.restore THEN
    BEGIN
        sh.x:=x;
        sh.y:=y;
        SaveArea(sh.a,x,y,sh.sx,sh.sy)
    END;
    PutImage(x,y,sh.p^,Mode)
END;


PROCEDURE HideShape(VAR sh:Shape);
BEGIN
    RestoreArea(sh.a)
END;


PROCEDURE MoveShape(VAR sh:Shape;x,y:INTEGER;Mode:WORD);
BEGIN
    RestoreArea(sh.a);
    sh.x:=x;
    sh.y:=y;
    SaveArea(sh.a,x,y,sh.sx,sh.sy);
    PutImage(x,y,sh.p^,Mode)
END;


FUNCTION StyleCorrect(style:WORD):REAL;
BEGIN
    CASE style OF
        0:StyleCorrect:=1;
        1:StyleCorrect:=200/266;
        2:StyleCorrect:=200/100;
        3:StyleCorrect:=200/263;
        4:StyleCorrect:=200/267;
        5:StyleCorrect:=200/378;
        6:StyleCorrect:=200/266
    END
END;


PROCEDURE SaveGraphParams;
BEGIN
    IF GraphSaveTop=maxGraphSave THEN
        GraphError(1);
    INC(GraphSaveTop);
    WITH GraphSaveList[GraphSaveTop] DO
    BEGIN
        GetTextSettings(TextSave);
        GetFillSettings(FillSave);
        GetLineSettings(LineSave);
        GetFillPattern(FillPatternSave);
        BkColor:=GetBkColor;
        Color:=GetColor
    END
END;


PROCEDURE RestoreGraphParams;
BEGIN
    IF GraphSaveTop=0 THEN
        GraphError(2);
    WITH GraphSaveList[GraphSaveTop] DO
    BEGIN
        SetTextJustify(TextSave.Horiz,TextSave.Vert);
        SetTextStyle(TextSave.Font,TextSave.Direction,TextSave.CharSize);
        SetLineStyle(LineSave.LineStyle,LineSave.Pattern,LineSave.Thickness);
        SetFillStyle(FillSave.Pattern,FillSave.Color);
        SetBkColor(BkColor);
        SetColor(Color);
        IF FillSave.Pattern=UserFill THEN
            SetFillPattern(FillPatternSave,FillSave.Color)
    END;
    DEC(GraphSaveTop)
END;


PROCEDURE MakeDithPalette;
VAR i,j,k:BYTE;
    thePal:ColorPalette256;
BEGIN
    FOR i:=0 TO 7 DO
        FOR j:=0 TO 7 DO
            FOR k:=0 TO 3 DO
            BEGIN
                thePal[i*32+j*4+k,0]:=i*9;
                thePal[i*32+j*4+k,1]:=j*9;
                thePal[i*32+j*4+k,2]:=k*21
            END;

    SetAllVGAPalette256(thePal)
END;


PROCEDURE MakeRedGreenPalette;
VAR i,j,k:BYTE;
    thePal:ColorPalette256;
BEGIN
    FOR i:=0 TO 15 DO
        FOR j:=0 TO 15 DO
        BEGIN
            thePal[i*16+j,0]:=i*4+2;
            thePal[i*16+j,1]:=j*4+2;
            thePal[i*16+j,2]:=0
        END;

    SetAllVGAPalette256(thePal)
END;


PROCEDURE MakeBlackWhitePalette;
VAR i:BYTE;
    thePal:ColorPalette256;
BEGIN
    FOR i:=0 TO 255 DO
    BEGIN
        thePal[i,0]:=i DIV 4;
        thePal[i,1]:=i DIV 4;
        thePal[i,2]:=i DIV 4
    END;
    SetAllVGAPalette256(thePal)
END;


PROCEDURE DithPutPixel(x,y:INTEGER;Red,Green,Blue:INTEGER);

CONST DithMat:ARRAY[0..3,0..3] OF BYTE=
          (( 0,12, 3,15),
           ( 8, 4,11, 7),
           ( 2,14, 1,13),
           (10, 6, 9, 5));

VAR RedMod,BlueMod,GreenMod:BYTE;
    m:BYTE;
BEGIN
    m:=DithMat[x AND 3,y AND 3];
    RedMod:=(Red AND $07) SHL 1;
    Red:=PRED(Red SHR 3);
    IF (RedMod>m) OR (Red<0) THEN
        INC(Red);
    GreenMod:=(Green AND $07) SHL 1;
    Green:=PRED(Green SHR 3);
    IF (GreenMod>m) OR (Green<0) THEN
        INC(Green);
    BlueMod:=Blue AND $0F;
    Blue:=PRED(Blue SHR 4);
    IF (BlueMod>m) OR (Blue<0) THEN
        INC(Blue);
    PutPixel(x,y,Red SHL 5 OR Green SHL 2 OR Blue)
END;


PROCEDURE RGPutPixel(x,y:INTEGER;Red,Green:INTEGER);
BEGIN
    PutPixel(x,y,(Red SHR 2) SHL 4+Green SHR 2)
END;


PROCEDURE SetDithColor(r,g,b:INTEGER);
BEGIN
    DithColor.r:=r;
    DithColor.g:=g;
    DithColor.b:=b
END;


PROCEDURE DithLine(x0,y0,x1,y1:INTEGER);
VAR dx,dy:INTEGER;
    i:INTEGER;
    dd:Real;
BEGIN
    dx:=x1-x0;
    dy:=y1-y0;
    WITH DithColor DO
        IF ABS(dx)>ABS(dy) THEN
        BEGIN
            dd:=dy/dx;
            IF dx>0 THEN
                FOR i:=x0 TO x1 DO
                    DithPutPixel(i,y0+ROUND(dd*(i-x0)),r,g,b)
            ELSE
                FOR i:=x1 TO x0 DO
                    DithPutPixel(i,y0+ROUND(dd*(i-x0)),r,g,b)
        END
        ELSE IF dy>0 THEN
        BEGIN
            dd:=dx/dy;
            IF dy>0 THEN
                FOR i:=y0 TO y1 DO
                    DithPutPixel(x0+ROUND(dd*(i-y0)),i,r,g,b)
            ELSE
                FOR i:=y1 TO y0 DO
                    DithPutPixel(x0+ROUND(dd*(i-y0)),i,r,g,b)
        END
END;


PROCEDURE OutTextXYNumb(x,y:INTEGER;s:String;nr:INTEGER);
VAR numStr:STRING;
BEGIN
    Str(nr,numStr);
    OutTextXY(x,y,s+numStr)
END;


BEGIN
    GraphSaveTop:=0;
    GrafDriver:=detect;
    GrafMode:=0;
    GraphR:=0
END.